import gzip
import logging
import zlib
from asyncio import ensure_future, get_event_loop
from asyncio.sslproto import SSLProtocol
from datetime import datetime

from . import constant
from .gelf import ChunkedMessage

logger = logging.getLogger(__name__)


class DefaultProtocol(object):
    def __init__(self, *args, **kwargs):
        raise NotImplementedError


class BaseLoggingProtocol(object):
    __slots__ = ['_sink', '_length', '_buffer', '_transport']
    PROTOCOL = DefaultProtocol

    def __init__(self, sink):
        self._sink = sink
        self._length = 0
        self._buffer = bytearray()
        self._transport = None

    def connection_made(self, transport):
        if hasattr(transport, 'sendto'):
            return
        else:
            self._transport = transport

    def data_received(self, data, addr=None):
        ensure_future(self._process_data(data, addr))

    def datagram_received(self, data, addr):
        # Append terminator to simplify parsing
        self.data_received(data + b'\x00', addr)

    def eof_received(self):
        pass

    def error_received(self, exc):
        self._close_with_error('Protocol error: {0}'.format(exc))

    def connection_lost(self, exc):
        pass

    def pause_writing(self):
        pass

    def resume_writing(self):
        pass

    def _close_with_error(self, message=None):
        if message:
            logger.error(message)
        if self._buffer:
            self._buffer.clear()
        if self._transport:
            self._transport.close()

    def _close_with_http(self):
        if self._buffer:
            self._buffer.clear()
        if self._transport:
            logger.debug('Sending HTTP response and closing connection')
            self._transport.write(b'HTTP/1.1 200 OK\r\nContent-Length: 0\r\nConnection: close\r\n\r\n')
            self._transport.close()

    def _get_non_transparent_framed_message(self):
        """
        Handle non-transparently-framed messages by searching for one of several possible terminators
        https://tools.ietf.org/html/rfc6587#section-3.4.2
        """
        for separator in constant.TERMS:
            message, sep, remainder = self._buffer.partition(bytes([separator]))
            if sep:
                if len(message) > constant.MAX_MESSAGE_LENGTH:
                    message = message[:constant.MAX_MESSAGE_LENGTH]
                self._buffer = remainder
                return message

    async def _process_data(self, data, addr=None):
        raise NotImplementedError


class BaseSecureLoggingProtocol(SSLProtocol):
    def __init__(self, sslcontext, *args, **kwargs):
        loop = get_event_loop()
        super(BaseSecureLoggingProtocol, self).__init__(
            loop=loop,
            app_protocol=self.PROTOCOL(*args, **kwargs),
            sslcontext=sslcontext,
            waiter=None,
            server_side=True
        )


class SyslogProtocol(BaseLoggingProtocol):
    async def _process_data(self, data, addr=None):
        if self._transport and self._transport.is_closing():
            return

        self._buffer.extend(data)
        while self._buffer:
            message = None

            if self._length or self._buffer[0] in constant.DIGITS:
                message = self._get_octet_counted_message()
            elif self._buffer[0] in constant.TERMS:
                del self._buffer[0]
            elif self._buffer[0] in constant.METHODS:
                self._close_with_http()
            else:
                message = self._get_non_transparent_framed_message()

            if message:
                addr = addr or self._transport.get_extra_info('peername')
                await self._sink.write(addr[0], bytes(message), datetime.now())
            else:
                return

    def _get_octet_counted_message(self):
        """
        Handle octet-counted messages by reading a space-separated length prefix,
        followed by the specified number of bytes. If insufficient bytes have been
        accumulated, remember the length and try again when we get more data.
        https://tools.ietf.org/html/rfc6587#section-3.4.1
        https://tools.ietf.org/html/rfc5425#section-4.3.1
        """
        if not self._length:
            length, sep, remainder = self._buffer.partition(b' ')
            if sep:
                try:
                    self._length = int(length)
                    if self._length < 0 or self._length > constant.MAX_MESSAGE_LENGTH:
                        raise ValueError
                    self._buffer = remainder
                except ValueError:
                    # Handle as random chunk of log noncompliant message that happens to start with numbers
                    return self._buffer
            else:
                return

        if len(self._buffer) >= self._length:
            message = self._buffer[:self._length]
            del self._buffer[:self._length]
            self._length = 0
            return message
        else:
            return


class GelfProtocol(BaseLoggingProtocol):
    async def _process_data(self, data, addr=None):
        if self._transport and self._transport.is_closing():
            return

        self._buffer.extend(data)
        while self._buffer:
            message = None

            if self._buffer[0] == constant.OPENBRACKET:
                message = self._get_non_transparent_framed_message()
            elif self._buffer[0:1] == constant.ZLIB_MAGIC:
                message = self._get_zlib_message()
            elif self._buffer[0:2] == constant.GZIP_MAGIC:
                message = self._get_gzip_message()
            elif self._buffer[0] in constant.TERMS:
                del self._buffer[0]
            elif self._buffer[0] in constant.METHODS:
                self._close_with_http()
            else:
                self._close_with_error('{0} unable to determine framing for message: {1}'.format(self.__class__.__name__, self._buffer))

            if message:
                addr = addr or self._transport.get_extra_info('peername')
                await self._sink.write(addr[0], bytes(message), datetime.now())
            else:
                return

    def _get_zlib_message(self):
        try:
            return zlib.decompress(self._buffer)
        except Exception:
            logger.error('ZLIB decompression failed', exc_info=True)
        finally:
            self._buffer.clear()

    def _get_gzip_message(self):
        try:
            return gzip.decompress(self._buffer)
        except Exception:
            logger.error('GZIP decompression failed', exc_info=True)
        finally:
            self._buffer.clear()


class SecureSyslogProtocol(BaseSecureLoggingProtocol):
    PROTOCOL = SyslogProtocol


class SecureGelfProtocol(BaseSecureLoggingProtocol):
    PROTOCOL = GelfProtocol


class DatagramSyslogProtocol(SyslogProtocol):
    pass


class DatagramGelfProtocol(GelfProtocol):
    __slots__ = ['_chunks']

    def __init__(self, *args, **kwargs):
        super(DatagramGelfProtocol, self).__init__(*args, **kwargs)
        self._chunks = dict()

    def datagram_received(self, data, addr):
        if data[0:2] == constant.GELF_MAGIC:
            data = self._process_chunk(data)
            if not data:
                return
        super(DatagramGelfProtocol, self).data_received(data, addr)

    # TODO - enforce 5 second chunk reassembly window
    def _process_chunk(self, data):
        chunk = ChunkedMessage(data)
        # Abuse the fact that ChunkedMessage hashes by ID
        if chunk in self._chunks:
            message = self._chunks[chunk].update(chunk)
        else:
            self._chunks[chunk] = chunk
            message = chunk.update()

        if message:
            del self._chunks[chunk]
        return message


# https://github.com/python/cpython/pull/480
def patch_pbo_29743():
    def _start_shutdown(self):
        if self._in_shutdown:
            return
        if self._in_handshake:
            self._abort()
        else:
            self._in_shutdown = True
            self._write_appdata(b'')

    SSLProtocol._start_shutdown = _start_shutdown


patch_pbo_29743()



import logging
import os
import time
from glob import glob
from multiprocessing import Lock
from tempfile import NamedTemporaryFile

from boto3 import Session
from botocore.config import Config

from .util import get_region

FLUSH_TIME = 60
MAX_REQUEST_SIZE = 1024 * 1024 * 4
MAX_RECORD_COUNT = 500
logger = logging.getLogger(__name__)
lock = Lock()


class EventSpool(object):
    __slots__ = ['delivery_stream', 'spool_dir', 'profile_name', 'config', 'flushed']
    PREFIX = 'firehose_event-'

    def __init__(self, delivery_stream, spool_dir, region_name=None, profile_name=None):
        self.delivery_stream = delivery_stream
        self.spool_dir = spool_dir
        self.profile_name = profile_name
        self.config = Config(retries={'max_attempts': 10}, region_name=get_region(region_name, profile_name))
        self.flushed = 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.flush()

    def write(self, record):
        with NamedTemporaryFile(prefix=self.PREFIX, dir=self.spool_dir, delete=False) as temp:
            logger.info('Writing {0} byte record to {1}'.format(len(record), temp.name))
            temp.write(record)
        if time.time() - self.flushed >= FLUSH_TIME:
            self.flush()

    def flush(self):
        if lock.acquire(block=False):
            try:
                while True:
                    record_files = glob(os.path.join(self.spool_dir, self.PREFIX) + '*')
                    request_kwargs = {'DeliveryStreamName': self.delivery_stream, 'Records': []}
                    request_size = 0
                    request_files = []

                    for path in record_files:
                        file_size = os.path.getsize(path)
                        if request_size + file_size <= MAX_REQUEST_SIZE and len(request_files) < MAX_RECORD_COUNT:
                            logger.debug('Including {0} ({1} bytes) in batch'.format(path, file_size))
                            request_size += file_size
                            request_files.append(path)
                        else:
                            break

                    for request_file in request_files:
                        try:
                            with open(request_file, 'rb') as fh:
                                request_kwargs['Records'].append({'Data': fh.read()})
                        except Exception:
                            logger.error('Failed to read file {0} from spool'.format(request_file), exc_info=True)
                            return

                    if request_kwargs['Records']:
                        logger.info('Batch has {0} bytes in {1} files'.format(request_size, len(request_files)))
                        try:
                            session = Session(profile_name=self.profile_name)
                            client = session.client('firehose', config=self.config)
                            response = client.put_record_batch(**request_kwargs)
                        except Exception:
                            logger.error('Firehose put_record_batch failed', exc_info=True)
                            return

                        for (i, status) in enumerate(response['RequestResponses']):
                            if 'RecordId' in status:
                                logger.debug('Firehose record succeeded: {0}'.format(request_files[i]))
                                try:
                                    os.unlink(request_files[i])
                                except Exception:
                                    logger.error('Failed to unlink successfully processed record: {0}'.format(request_files[i]))
                            else:
                                logger.warning('Firehose record failed: [{ErrorCode}] {ErrorMessage}'.format(**status))
                    else:
                        logger.debug('Batch is empty')
                        return
            finally:
                self.flushed = time.time()
                lock.release()



import logging

import boto3
import botocore

import ujson as json

logger = logging.getLogger(__name__)


def get_instance_region():
    data = {}
    fetcher = botocore.utils.InstanceMetadataFetcher()

    try:
        r = fetcher._get_request('http://169.254.169.254/latest/dynamic/instance-identity/document', fetcher._timeout, fetcher._num_attempts)
        if r.content:
            val = r.content.decode('utf-8')
            if val[0] == '{':
                data = json.loads(val)
    except botocore.utils._RetriesExceededError:
        logger.debug("Max number of attempts exceeded ({0}) when attempting to retrieve data from metadata service.".format(fetcher._num_attempts))

    return data.get('region', None)


def get_region(region_name=None, profile_name=None):
    region = region_name or boto3.Session(profile_name=profile_name).region_name or get_instance_region()
    if not region:
        raise botocore.exceptions.NoRegionError
    return region



import logging
import ssl
from asyncio import get_event_loop

from .protocol import (DatagramGelfProtocol, DatagramSyslogProtocol,
                       DefaultProtocol, GelfProtocol, SecureGelfProtocol,
                       SecureSyslogProtocol, SyslogProtocol)

logger = logging.getLogger(__name__)


class BaseServer(object):
    __slots__ = ['_host', '_port', '_loop', '_args']
    PROTOCOL = DefaultProtocol

    def __init__(self, host, port):
        logger.info('Starting {0} on {1}:{2}'.format(self.__class__.__name__, host, port))
        self._host = host
        self._port = port
        self._loop = get_event_loop()
        self._args = {}

    def _protocol_factory(self, sink):
        return lambda: self.PROTOCOL(sink=sink, **self._args)

    async def start_server(self, sink):
        return await self._loop.create_server(self._protocol_factory(sink), self._host, self._port)


class SecureServer(BaseServer):
    def __init__(self, certfile, keyfile, password=None, *args, **kwargs):
        super(SecureServer, self).__init__(*args, **kwargs)
        logger.info('{0} using cert from {1} and key from {2}'.format(self.__class__.__name__, certfile, keyfile))
        ctx = ssl.SSLContext(ssl.PROTOCOL_TLS)
        ctx.load_cert_chain(certfile, keyfile, password)
        self._args['sslcontext'] = ctx


class DatagramServer(BaseServer):
    async def start_server(self, sink):
        return await self._loop.create_datagram_endpoint(self._protocol_factory(sink), local_addr=(self._host, self._port))


class SyslogServer(BaseServer):
    PROTOCOL = SyslogProtocol


class SecureSyslogServer(SecureServer):
    PROTOCOL = SecureSyslogProtocol


class DatagramSyslogServer(DatagramServer):
    PROTOCOL = DatagramSyslogProtocol


class GelfServer(BaseServer):
    PROTOCOL = GelfProtocol


class SecureGelfServer(SecureServer):
    PROTOCOL = SecureGelfProtocol


class DatagramGelfServer(DatagramServer):
    PROTOCOL = DatagramGelfProtocol



import logging
import math
import time
from asyncio import get_event_loop
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor
from gzip import compress

from boto3 import Session

import ujson as json

from . import constant

logger = logging.getLogger(__name__)


class MessageSink(object):
    __slots__ = ['spool', 'loop', 'executor', 'size', 'count', 'messages', 'flushed', 'message_class', 'account']

    def __init__(self, spool, message_class):
        self.spool = spool
        self.message_class = message_class
        self.loop = get_event_loop()
        self.executor = ProcessPoolExecutor()
        self.executor._start_queue_management_thread()
        self._schedule_flush()
        self.clear()
        self.account = '000000000000'
        try:
            session = Session(profile_name=spool.profile_name)
            client = session.client('sts', config=spool.config)
            self.account = client.get_caller_identity()['Account']
        except Exception:
            logger.warn('Unable to determine AWS Account ID; using default value.')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.flush()

    async def write(self, source, message, timestamp):
        self.messages[source].append((message, timestamp))
        self.size += len(message)
        self.count += 1
        if self.size > constant.FLUSH_SIZE:
            await self.flush_async()
        return len(message)

    def clear(self):
        self.size = 0
        self.count = 0
        self.messages = defaultdict(list)
        self.flushed = time.time()

    async def flush_async(self):
        self.loop.run_in_executor(self.executor, self._spool_messages, self.spool, self.messages, self.size, self.message_class, self.account)
        self.clear()

    def flush(self):
        self._spool_messages(self.spool, self.messages, self.size, self.message_class, self.account)
        self.clear()

    def _schedule_flush(self):
        self.loop.call_later(constant.TIMER_INTERVAL, self._flush_timer)

    def _flush_timer(self):
        logger.debug('flush timer: messages={0} size={1} age={2}'.format(self.count, self.size, time.time() - self.flushed))
        if self.messages and time.time() - self.flushed >= constant.FLUSH_TIME:
            self.loop.create_task(self.flush_async())
        self._schedule_flush()

    @classmethod
    def _spool_messages(cls, spool, messages, size, message_class, account):
        for i_source, i_messages in messages.items():
            events = message_class.create_events(i_source, i_messages)
            record = cls._prepare_record(i_source, events, message_class.name, account)
            compressed_record = cls._compress_record(record)
            logger.debug('Events for {0} compressed from {1} to {2} bytes (with JSON framing)'.format(i_source, size, len(compressed_record)))

            if len(compressed_record) > constant.MAX_RECORD_SIZE:
                # This approach naievely hopes that splitting a record into even parts will put it
                # below the max record size. Further tuning may be required.
                split_count = math.ceil(len(compressed_record) / constant.MAX_RECORD_SIZE)
                logger.warning('Compressed record size of {0} bytes exceeds maximum Firehose record size of {1} bytes; splitting into {2} records'.format(
                    len(compressed_record),
                    constant.MAX_RECORD_SIZE,
                    split_count
                ))
                start = 0
                size = int(len(record['logEvents']) / split_count)
                while start < len(record['logEvents']):
                    record_part = cls._prepare_record(i_source, record['logEvents'][start:start+size], message_class.name, account)
                    compressed_record = cls._compress_record(record_part)
                    spool.write(compressed_record)
                    start += size
            else:
                spool.write(compressed_record)

    @classmethod
    def _prepare_record(cls, source, events, class_name, account):
        return {
            'owner': account,
            'logGroup': class_name,
            'logStream': source,
            'subscriptionFilters': [class_name],
            'messageType': 'DATA_MESSAGE',
            'logEvents': events,
        }

    @classmethod
    def _compress_record(cls, record):
        return compress(json.dumps(record).encode())



from struct import Struct

MAX_MESSAGE_LENGTH = 1024 * 16  # Maximum supported message length
DIGITS = bytearray(i for i in range(0x30, 0x3A))  # First char of octet-counted syslog message
TERMS = bytearray([0x00, 0x0A, 0x0D])  # Framed message terminators
METHODS = bytearray([0x47, 0x48])  # First char of common HTTP methods
ZLIB_MAGIC = b'\x78'  # ZLIB magic
GZIP_MAGIC = b'\x1F\x8B'  # GZIP magic
GELF_MAGIC = b'\x1E\x0F'  # Chunked GELF header magic
GELF_HEADER = Struct('!HqBB')  # Chunked GELF header format
FLUSH_TIME = 60  # Sink buffer max message age
FLUSH_SIZE = 1024 * 1024 * 4  # Sink buffer max size
MAX_RECORD_SIZE = 1024 * 1000  # Maximum Firehose record size
TIMER_INTERVAL = 10  # Sink buffer autoflush check interval
PROXY10_SEP = b'\x20'  # Proxy Protocol v1.0 separator
PROXY10_TERM = b'\x0D\x0A'  # Proxy Protocol v1.0 terminator
PROXY10_MAGIC = b'PROXY\x20'  # Proxy Protocol v1.0 header
PROXY20_MAGIC = b'\x0D\x0A\x0D\x0A\x00\x0D\x0AQUIT\x0A'  # Proxy Protocol v2.0 header
PROXY20_COMMANDS = {0x00: 'local',
                    0x01: 'proxy',
                    }
PROXY20_FAMILIES = {0x10: 'inet',
                    0x20: 'inet6',
                    0x30: 'unix',
                    }
PROXY20_PROTOCOLS = {0x01: 'stream',
                     0x02: 'datagram',
                     }
PROXY20_TLV_TYPES = {0x01: 'PP2_TYPE_ALPN',
                     0x02: 'PP2_TYPE_AUTHORITY',
                     0x03: 'PP2_TYPE_CRC32C',
                     0x04: 'PP2_TYPE_NOOP',
                     0x20: 'PP2_TYPE_SSL',
                     0x26: 'PP2_TYPE_NETNS',
                     0xEA: 'PP2_TYPE_AWS',
                     }



import logging
import socket
import struct
from asyncio import get_event_loop, protocols, transports

from . import constant

logger = logging.getLogger(__name__)


class _ProxyProtocolTransport(transports._FlowControlMixin, transports.Transport):
    """
    Basic wrapped transport inspired by _SSLProtocolTransport.

    Delegates almost everything to the underlying transport, except get_extra_info
    which will substitute Proxy Protocol peerinfo and TLVs if available.
    """
    def __init__(self, loop, proxy_protocol):
        self._loop = loop
        self._proxy_protocol = proxy_protocol
        self._closed = False

    def get_extra_info(self, name, default=None):
        return self._proxy_protocol._get_extra_info(name, default)

    def set_protocol(self, protocol):
        self._proxy_protocol.app_protocol = protocol

    def get_protocol(self):
        return self._proxy_protocol._app_protocol

    def is_closing(self):
        return self._closed

    def close(self):
        self._closed = True

    def __del__(self):
        if not self._closed:
            self.close()

    def is_reading(self):
        return self._proxy_protocol._transport.is_reading()

    def pause_reading(self):
        self._proxy_protocol._transport.pause_reading()

    def resume_reading(self):
        self._proxy_protocol._transport._resume_reading()

    def set_write_buffer_limits(self, high=None, low=None):
        self._proxy_protocol._transport.set_write_buffer_limits(high, low)

    def get_write_buffer_size(self):
        return self._proxy_protocol._transport.get_write_buffer_size()

    @property
    def _protocol_paused(self):
        return self._proxy_protocol._transport._protocol_paused

    def write(self, data):
        self._proxy_protocol._transport.write(data)

    def can_write_eof(self):
        return False

    def abort(self):
        return self._proxy_protocol._transport.abort()


class _BaseProxyProtocol(protocols.Protocol):
    """
    Protocol wrapper inspired by SSLProtocol.

    Once a Proxy Protocol header has been found, the wrapper gets out of the way
    and passes data directly to the app protocol.
    """
    def __init__(self, loop, app_protocol):
        self._loop = loop
        self._app_protocol = app_protocol
        self._app_transport = _ProxyProtocolTransport(self._loop, self)
        self._session_established = False
        self._transport = None
        self._buffer = bytearray()
        self._extra = dict()

    def connection_made(self, transport):
        self._transport = transport

    def connection_lost(self, exc):
        if self._session_established:
            self._session_established = False
            self._loop.call_soon(self._app_protocol.connection_lost, exc)
        self._transport = None
        self._app_transport = None

    def pause_writing(self):
        self._app_protocol.pause_writing()

    def resume_writing(self):
        self._app_protocol.resume_writing()

    def data_received(self, data):
        if self._session_established:
            self._loop.call_soon(self._app_protocol.data_received, data)
        else:
            self._buffer.extend(data)
            self._parse_proxy_protocol_header()

    def eof_received(self):
        self._transport.close()

    def _get_extra_info(self, name, default=None):
        if name in self._extra:
            return self._extra[name]
        elif self._transport is not None:
            self._transport.get_extra_info(name, default)
        else:
            return default

    def _parse_proxy_protocol_header(self):
        if self._buffer.startswith(constant.PROXY10_MAGIC):
            self._parse_proxy10()
        elif self._buffer.startswith(constant.PROXY20_MAGIC):
            self._parse_proxy20()
        elif len(self._buffer) > len(constant.PROXY20_MAGIC):
            self._close_with_error('PROXY protocol error: invalid header')
        else:
            logger.debug('Waiting for more data...')

    def _parse_proxy10(self):
        header, sep, payload = self._buffer.partition(constant.PROXY10_TERM)
        if sep:
            try:
                header, proto, src_addr, dst_addr, src_port, dst_port = header.split(constant.PROXY10_SEP)
                self._extra['peername'] = (src_addr.decode(), int(src_port))
                self._extra['sockname'] = (dst_addr.decode(), int(dst_port))
                self._start_session(payload)
            except Exception:
                self._close_with_error('PROXY protocol error: invalid header')

    def _parse_proxy20(self):
        if len(self._buffer) < 16:
            self._close_with_error('PROXY protocol error: invalid header')

        version = self._buffer[12] & 0xF0
        command = self._buffer[12] & 0x0F
        family = self._buffer[13] & 0xF0
        protocol = self._buffer[13] * 0x0F  # NOQA F841
        addr_len = struct.unpack('!H', self._buffer[14:16])[0]

        tlv_end = addr_len + 16
        if len(self._buffer) < tlv_end:
            # Don't have entire header yet; wait for more to come in
            return

        if version != 0x20:
            self._close_with_error('PROXY protocol error: invalid version')

        if command not in constant.PROXY20_COMMANDS:
            self._close_with_error('PROXY protocol error: invalid command')

        if constant.PROXY20_COMMANDS[command] == 'proxy':
            if family not in constant.PROXY20_FAMILIES:
                self._close_with_error('PROXY protocol error: invalid address family')

            if constant.PROXY20_FAMILIES[family] == 'inet':
                src_ip, dst_ip, src_port, dst_port = struct.unpack('!4s4sHH', self._buffer[16:28])
                self._extra['peername'] = (socket.inet_ntop(socket.AF_INET, src_ip), src_port)
                self._extra['sockname'] = (socket.inet_ntop(socket.AF_INET, dst_ip), dst_port)
                self._parse_tlv_data(28, tlv_end)
            elif constant.PROXY20_FAMILIES[family] == 'inet6':
                src_ip, dst_ip, src_port, dst_port = struct.unpack('!16s16sHH', self._buffer[16:52])
                self._extra['peername'] = (socket.inet_ntop(socket.AF_INET6, src_ip), src_port)
                self._extra['sockname'] = (socket.inet_ntop(socket.AF_INET6, dst_ip), dst_port)
                self._parse_tlv_data(52, tlv_end)
            elif constant.PROXY20_FAMILIES[family] == 'unix':
                src_addr, dst_addr = struct.unpack('!108s108s', self._buffer[16:232])
                self._extra['peername'] = (src_addr.rstrip(b'\x00'),)
                self._extra['sockname'] = (dst_addr.rstrip(b'\x00'),)
                self._parse_tlv_data(232, tlv_end)

            self._start_session(self._buffer[tlv_end:])

    def _parse_tlv_data(self, start, end):
        try:
            while start < end and start < len(self._buffer):
                tlv_type = self._buffer[start]
                tlv_len = struct.unpack('!H', self._buffer[start+1:start+3])[0]
                if tlv_type in constant.PROXY20_TLV_TYPES:
                    tlv_type = constant.PROXY20_TLV_TYPES[tlv_type]
                    if tlv_type != 'PP2_TYPE_NOOP':
                        tlv_value = bytes(self._buffer[start+3:start+3+tlv_len])
                        self._extra[tlv_type] = tlv_value
                else:
                    logger.warn('Unknown TLV type 0x{0:x}'.format(tlv_type))
                start += 3 + tlv_len
        except Exception:
            self._close_with_error('PROXY protocol error: Invalid TLV data')

    def _start_session(self, data):
        self._buffer.clear()
        self._app_protocol.connection_made(self._app_transport)
        self._session_established = True
        self.data_received(data)

    def _close_with_error(self, message=None):
        if message:
            logger.error(message)
        if self._buffer:
            self._buffer.clear()
        if self._transport:
            self._transport.close()


def wrap(cls):
    """
    Wrap a Server class with Proxy Prococol support
    The base Server and Protocol will be used after Proxy protocol headers have been validated.
    """
    loop = get_event_loop()

    class Protocol(_BaseProxyProtocol):
        def __init__(self, *args, **kwargs):
            super(Protocol, self).__init__(loop=loop, app_protocol=cls.PROTOCOL(*args, **kwargs))

    class Server(cls):
        PROTOCOL = Protocol

        def __init__(self, *args, **kwargs):
            super(Server, self).__init__(*args, **kwargs)

    Protocol.__name__ = 'Proxy'+cls.PROTOCOL.__name__
    Server.__name__ = 'Proxy'+cls.__name__

    return Server



import click

from . import core
from pkg_resources import get_distribution


def _print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo(get_distribution(__name__.split('.')[0]).version)
    ctx.exit()


@click.group()
@click.option(
    '--version',
    is_flag=True,
    callback=_print_version,
    expose_value=False,
    is_eager=True,
    help='Show current tool version.'
)
def cli():
    pass


cli.add_command(core.listen)

if __name__ == '__main__':
    cli()



import logging
import re
from datetime import datetime, timedelta

from libuuid import uuid4

import ujson as json

logger = logging.getLogger(__name__)
epoch = datetime.utcfromtimestamp(0)
prio_pattern = re.compile(r"""^<(?P<prio>\d{1,3})>
                                (?P<content>.*)""", re.VERBOSE)
syslog_pattern = re.compile(r"""^<(?P<prio>\d{1,3})>
                               (?:(?P<rfc5424>1\ )?
                                  (?P<timestamp>\S{20,38}|...\ ..\ ..:..:..(\ 20\d\d)?|-))\ ?
                                  (?P<hostname>\S+)*\ ?
                                  (?P<content>.*)""", re.VERBOSE | re.IGNORECASE)


def parse_rfc3164_timestamp(timestamp):
    if len(timestamp) <= 15:
        timestamp = '{0} {1}'.format(timestamp, datetime.today().year)
    parsed_time = datetime.strptime(timestamp, '%b %d %H:%M:%S %Y')

    # If the timetamp is more than 2 days in the future, it's probably from last year
    if (parsed_time - datetime.now()).days > 2:
        parsed_time = parsed_time.replace(year=parsed_time.year - 1)

    return parsed_time


def parse_rfc5424_timestamp(timestamp):
    if timestamp[-1] == 'Z':
        return parse_rfc5424_date(timestamp[:-1])
    else:
        value = parse_rfc5424_date(timestamp[:-6])
        hours = int(timestamp[-5:-3])
        mins = int(timestamp[-2:])
        sign = 1 if timestamp[-6] == '-' else -1
        return value + timedelta(seconds=sign*(hours*3600+mins*60))


def parse_rfc5424_date(timestamp):
    if '.' in timestamp:
        return datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S.%f")
    else:
        return datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S")


def format_rfc5424_date(timestamp):
    return datetime.strftime(timestamp, "%Y-%m-%dT%H:%M:%S.%f")


def assign_uuid(message, timestamp):
    if isinstance(timestamp, datetime):
        timestamp = (timestamp - epoch).total_seconds()

    return {'id': str(uuid4()),
            'message': message,
            'timestamp': timestamp,
            }


class BaseMessage(object):
    name = 'base'

    @classmethod
    def create_events(cls, source, messages):
        for message, recv_ts in messages:
            yield cls.create_event(source, message.decode(), recv_ts)

    @classmethod
    def create_event(cls, source, message, recv_ts):
        raise NotImplementedError


class GelfMessage(BaseMessage):
    name = 'gelf_message'

    @classmethod
    def create_event(cls, source, message, recv_ts):
        try:
            timestamp = json.loads(message).get('timestamp', recv_ts)
        except Exception:
            timestamp = recv_ts

        return assign_uuid(message, timestamp)


class SyslogMessage(BaseMessage):
    name = 'syslog_message'

    @classmethod
    def create_event(cls, source, message, recv_ts):
        parts = cls.get_message_header(source, message)
        try:
            timestamp = parse_rfc5424_timestamp(parts['timestamp']) if parts['rfc5424'] else parse_rfc3164_timestamp(parts['timestamp'])
        except Exception:
            timestamp = None

        if not parts['prio']:
            parts['prio'] = '13'

        if not isinstance(timestamp, datetime):
            timestamp = recv_ts
            message = '<{0}>1 {1} {2} {3}'.format(parts['prio'], format_rfc5424_date(timestamp), source, parts['content'])

        return assign_uuid(message, timestamp)

    @classmethod
    def get_message_header(cls, source, message):
        for pattern in [syslog_pattern, prio_pattern]:
            match = pattern.match(message)
            if match:
                return match.groupdict()
        return {
            'prio': None,
            'rfc5242': None,
            'timestamp': None,
            'hostname': None,
            'content': message,
            }



# -*- coding: utf-8 -*-
from os import chdir
from os.path import abspath, dirname

from setuptools import find_packages, setup

chdir(dirname(abspath(__file__)))

version = {}

with open('README.rst') as f:
    readme = f.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name='kinesyslog',
    version_command=('git describe --tags --dirty', 'pep440-git-full'),
    description='Syslog and GELF relay to AWS Kinesis Firehose. Supports UDP, TCP, and TLS; RFC3164, RFC5424, RFC5425, RFC6587, GELF v1.1.',
    long_description=readme,
    author='Brandon Davidson',
    author_email='brad@oatmail.org',
    url='https://github.com/brandond/kinesyslog',
    license='Apache',
    packages=find_packages(exclude=('docs')),
    entry_points={
        'console_scripts': ['kinesyslog=kinesyslog.commands:cli']
    },
    include_package_data=True,
    install_requires=requirements,
    python_requires='>=3.5',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: Log Analysis',
        'Topic :: System :: Logging',
    ],
    extras_require={
        'dev': [
            'setuptools-version-command',
        ]
    },
)



import logging
import signal
from asyncio import CancelledError, Task, gather, get_event_loop
from functools import partial
from tempfile import gettempdir

import click

from . import proxy
from .message import GelfMessage, SyslogMessage
from .server import (DatagramGelfServer, DatagramSyslogServer, GelfServer,
                     SecureGelfServer, SecureSyslogServer, SyslogServer)
from .sink import MessageSink
from .spool import EventSpool


def shutdown_exception_handler(loop, context):
    if "exception" not in context or not isinstance(context["exception"], CancelledError):
        loop.default_exception_handler(context)


@click.option(
    '--debug',
    is_flag=True,
    help='Enable debug logging to STDERR.',
)
@click.option(
    '--gelf',
    is_flag=True,
    help='Listen for messages in Graylog Extended Log Format (GELF) instead of Syslog.',
)
@click.option(
    '--profile',
    type=str,
    help='Use a specific profile from your credential file.',
)
@click.option(
    '--region',
    type=str,
    help='The region to use. Overrides config/env settings.',
)
@click.option(
    '--spool-dir',
    type=click.Path(exists=True, writable=True, file_okay=False, resolve_path=True),
    help='Spool directory for compressed records prior to upload.',
    default=gettempdir(),
    show_default=True,
)
@click.option(
    '--proxy-protocol',
    is_flag=True,
    help='Enable Proxy Protocol v1/v2 support for TCP and TLS listeners.',
)
@click.option(
    '--udp-port',
    type=int,
    help='Bind port for UDP listener; 0 to disable.',
    default=0,
    show_default=True,
    multiple=True,
)
@click.option(
    '--tcp-port',
    type=int,
    help='Bind port for TCP listener; 0 to disable.',
    default=0,
    show_default=True,
    multiple=True,
)
@click.option(
    '--key',
    type=click.Path(exists=True, readable=True, dir_okay=False, resolve_path=True),
    help='Private key file for TLS listener.',
    default='localhost.key',
    show_default=True,
)
@click.option(
    '--cert',
    type=click.Path(exists=True, readable=True, dir_okay=False, resolve_path=True),
    help='Certificate file for TLS listener.',
    default='localhost.crt',
    show_default=True,
)
@click.option(
    '--port',
    type=int,
    help='Bind port for TLS listener; 0 to disable.',
    default=6514,
    show_default=True,
    multiple=True,
)
@click.option(
    '--address',
    type=str,
    help='Bind address.',
    default='0.0.0.0',
    show_default=True,
)
@click.option(
    '--stream',
    type=str,
    help='Kinesis Firehose Delivery Stream Name.',
    required=True,
)
@click.command(short_help='List for incoming Syslog messages and submit to Kinesis Firehose')
def listen(**args):
    logging.basicConfig(level='INFO', format='%(asctime)-15s %(levelname)s:%(name)s %(message)s')
    loop = get_event_loop()
    loop.set_exception_handler(shutdown_exception_handler)

    if args.get('gelf', False):
        message_class = GelfMessage
        TLS = SecureGelfServer
        TCP = GelfServer
        UDP = DatagramGelfServer
    else:
        message_class = SyslogMessage
        TLS = SecureSyslogServer
        TCP = SyslogServer
        UDP = DatagramSyslogServer

    if args.get('proxy_protocol', False):
        TLS = proxy.wrap(TLS)
        TCP = proxy.wrap(TCP)

    if args.get('debug', False):
        logging.getLogger('kinesyslog').setLevel('DEBUG')
        logging.getLogger('asyncio').setLevel('INFO')
        loop.set_debug(True)
    else:
        logging.getLogger('botocore').setLevel('ERROR')

    servers = []
    try:
        if args.get('port', None):
            for port in args['port']:
                servers.append(TLS(host=args['address'], port=port, certfile=args['cert'], keyfile=args['key']))
        if args.get('tcp_port', None):
            for port in args['tcp_port']:
                servers.append(TCP(host=args['address'], port=port))
        if args.get('udp_port', None):
            for port in args['udp_port']:
                servers.append(UDP(host=args['address'], port=port))
    except Exception:
        logging.error('Failed to start server', exc_info=True)

    if not servers:
        return

    for signame in ('SIGINT', 'SIGTERM'):
        loop.add_signal_handler(getattr(signal, signame), partial(loop.stop))

    try:
        with EventSpool(delivery_stream=args['stream'], spool_dir=args['spool_dir']) as spool:
            with MessageSink(spool=spool, message_class=message_class) as sink:
                for server in servers:
                    loop.run_until_complete(server.start_server(sink=sink))
                loop.run_forever()
    except KeyboardInterrupt:
        tasks = gather(*Task.all_tasks(loop=loop), loop=loop, return_exceptions=True)
        tasks.add_done_callback(partial(loop.stop))
        tasks.cancel()
        while not tasks.done() and not loop.is_closed():
            loop.run_forever()
    finally:
        loop.close()



import logging
import time
from collections import namedtuple

from . import constant

Header = namedtuple('Header', 'magic id number count')
logger = logging.getLogger(__name__)


class ChunkedMessage(object):
    __slots__ = ['header', 'payload', 'parts', 'found', 'seen']

    def __init__(self, data):
        self.header = self.parse_header(data)
        self.payload = data[12:]
        self.parts = None
        self.found = 0
        self.seen = 0
        assert self.header.count <= 128

    def __hash__(self):
        return self.header.id

    def __eq__(self, other):
        return self.header.id == other.header.id

    def __ne__(self, other):
        return self.header.id != other.header.id

    def is_complete(self):
        return self.header.count == self.found

    def update(self, other=None):
        if other:
            assert other.header.id == self.header.id
        else:
            other = self
        if not self.parts:
            self.parts = [None] * self.header.count
        if not self.parts[other.header.number]:
            self.parts[other.header.number] = other.payload
            self.seen = time.time()
            self.found += 1
        if self.is_complete():
            return b''.join(self.parts)

    @classmethod
    def parse_header(cls, data):
        return Header(*constant.GELF_HEADER.unpack(data[0:12]))
